import json
import os

from pathlib import Path
from time import strftime, localtime
from typing import Dict

import conf.defaults
from model_server.model import Model
from model_server.share import SharedImageDirectory
from model_server.workflow import WorkflowRunRecord

def create_manifest_json():
    pass

class Session(object):
    """
    Singleton class for a server session that persists data between API calls
    """
    # inbound = SharedImageDirectory(conf.defaults.paths['inbound_images'])
    # outbound = SharedImageDirectory(conf.defaults.paths['outbound_images'])
    # where_records = conf.defaults.paths['logs']

    def __new__(cls):
        if not hasattr(cls, 'instance'):
            cls.instance = super(Session, cls).__new__(cls)
        return cls.instance

    def __init__(self, root: str = None):
        print('Initializing session')
        self.models = {} # model_id : model object
        self.manifest = [] # paths to data as well as other metadata from each inference run
        self.paths = self.make_paths(root)
        # self.session_id = self.create_session_id(self.paths['logs'])
        # self.session_id = self.create_session_id(self.paths)
        self.session_log = self.paths['logs'] / f'session.log'
        self.log_event('Initialized session')
        self.manifest_json = self.paths['logs'] / f'manifest.json'
        open(self.manifest_json, 'w').close() # instantiate empty json file

    def get_paths(self):
        return self.paths

    @staticmethod
    def make_paths(root: str = None) -> dict:
        """
        Set paths where images, logs, etc. are located in this session
        :param root: absolute path to top-level directory
        :return: dictionary of session paths
        """
        if root is None:
            root_path = Path(conf.defaults.root)
        else:
            root_path = Path(root)
        sid = Session.create_session_id(root_path)
        paths = {'root': root_path}
        for pk in ['inbound_images', 'outbound_images', 'logs']:
            pa = root_path / sid / conf.defaults.subdirectories[pk]
            paths[pk] = pa
            try:
                pa.mkdir(parents=True, exist_ok=True)
            except Exception:
                raise CouldNotCreateDirectory(f'Could not create directory: {pa}')
        return paths

    @staticmethod
    def create_session_id(look_where: Path) -> str:
        """
        Autogenerate a session ID by incrementing from a list of log files.
        """
        yyyymmdd = strftime('%Y%m%d', localtime())
        idx = 0
        while os.path.exists(look_where / f'{yyyymmdd}-{idx:04d}'):
            idx += 1
        return f'{yyyymmdd}-{idx:04d}'

    def log_event(self, event: str):
        """
        Write an event string to this session's log file.
        """
        timestamp = strftime('%m/%d/%Y, %H:%M:%S', localtime())
        with open(self.session_log, 'a') as fh:
            fh.write(f'{timestamp} -- {event}\n')

    def record_workflow_run(self, record: WorkflowRunRecord or None):
        """
        Append a JSON describing inference data to this session's manifest
        """
        self.log_event(f'Ran model {record.model_id} on {record.input_filepath} to infer {record.output_filepath}')
        with open(self.manifest_json, 'w+') as fh:
            json.dump(record.dict(), fh)


    def load_model(self, ModelClass: Model, params: Dict[str, str] = None) ->  dict:
        """
        Load an instance of a given model class and attach to this session's model registry
        :param ModelClass: subclass of Model
        :param params: optional parameters that are passed to the model's construct
        :return: dictionary that describes all currently loaded models
        """
        mi = ModelClass(params=params)
        assert mi.loaded, f'Error loading instance of {ModelClass.__name__}'
        ii = 0

        def mid(i):
            return f'{ModelClass.__name__}_{i:02d}'

        while mid(ii) in self.models.keys():
            ii += 1

        key = mid(ii)
        self.models[key] = {
            'object': mi,
            'params': params
        }
        self.log_event(f'Loaded model {key}')
        return key

    def describe_loaded_models(self) -> dict:
        return {
            k: {
                'class': self.models[k]['object'].__class__.__name__,
                'params': self.models[k]['params'],
            }
            for k in self.models.keys()
        }

    def find_param_in_loaded_models(self, key: str, value: str) -> dict:
        """
        Returns first instance of loaded model where key and value match with .params field, or None
        """
        models = self.describe_loaded_models()
        for mid, det in models.items():
            if det.get('params').get(key) == value:
                return {mid: det}
        return None

    def restart(self, **kwargs):
        self.__init__(**kwargs)

class Error(Exception):
    pass

class InferenceRecordError(Error):
    pass

class CouldNotInstantiateModelError(Error):
    pass

class CouldNotCreateDirectory(Error):
    pass